import numpy as np


def distance_matrix(matrix_a: np.ndarray, matrix_b: np.ndarray) -> np.ndarray:
    """Calculate the distance matrix between two matrices.

    Args:
        matrix_a (np.ndarray): Matrix of shape (n, d).
        matrix_b (np.ndarray): Matrix of shape (m, d).

    Returns:
        np.ndarray: The distance matrix, of shape (n, m)
    """
    # sqrt((A - B)^2) = sqrt(A^2 + B^2 - 2AB)

    matrix_a_2 = np.sum(np.square(matrix_a), axis=1)
    matrix_b_2 = np.sum(np.square(matrix_b), axis=1)
    a_b = np.dot(matrix_a, matrix_b.T)
    return np.sqrt(matrix_a_2[:, np.newaxis] + matrix_b_2 - 2 * a_b)


def knn_predict(dists: np.ndarray, labels_train: np.ndarray, k: int) -> np.ndarray:
    """Generate predictions based on the distance matrix between the train and test datasets.

    Args:
        dists (np.ndarray): The distance matrix (rows: test datapoints, columns: train datapoints), of size (n, m).
        labels_train (np.ndarray): The labels of the training data, of size (m,).
        k (int): The number of neighbors to use.

    Returns:
        np.ndarray: The predicted labels, of size (n,).
    """
    # for each row, get the indices of the k smallest values, then get the labels of those indices, then get the most common label
    # argpartition sorts the first k elements of each row (axis=1) and returns the sorted indices (more efficient than argsort)
    indices = np.argpartition(dists, k, axis=1)[:, :k]

    # get the labels of the indices
    nearest_labels = labels_train[indices]

    # get the most common label
    predicted_labels = np.array([np.bincount(row).argmax() for row in nearest_labels])

    return predicted_labels


def evaluate_knn(
    data_train: np.ndarray,
    labels_train: np.ndarray,
    data_test: np.ndarray,
    labels_test: np.ndarray,
    k: int,
) -> float:
    """Evaluate the accuracy of k-nearest neighbors on the given dataset.

    Args:
        data_train (np.ndarray): The training data.
        labels_train (np.ndarray): The training data labels.
        data_test (np.ndarray): The test data.
        labels_test (np.ndarray): The test data labels.
        k (int): The number of neighbors to use.

    Returns:
        float: The accuracy of the model (between 0 and 1).
    """
    dists = distance_matrix(data_test, data_train)
    predicted_labels = knn_predict(dists, labels_train, k)
    # Get array of ones and zeros and average over it
    accuracy = (predicted_labels == labels_test).mean()
    return accuracy
