import pickle

import numpy as np


def read_cifar_batch(path: str) -> (np.ndarray, np.ndarray):
    """Reads the batch of CIFAR-10 data corresponding to the given path.

    Args:
        path (str): The path to the batch file.

    Returns:
        (np.ndarray, np.ndarray): A tuple containing the data and labels contained in the batch file.
    """
    with open(path, "rb") as file:
        dict_ = pickle.load(file, encoding="bytes")
    raw_data = dict_[b"data"]
    data_array = np.array(raw_data, dtype=np.float32)

    raw_labels = dict_[b"labels"]
    labels_array = np.array(raw_labels, dtype=np.int64)

    return data_array, labels_array


def read_cifar(path: str) -> (np.ndarray, np.ndarray):
    """Reads the CIFAR-10 dataset from the given path. The path must contain files data_batch_1 to 5, and test_batch.

    Args:
        path (str): The path to the folder containing the batches.

    Returns:
        (np.ndarray, np.ndarray): A tuple containing the dataset's data and labels. The data also includes the test data.
    """
    batch_filenames = [f"{path}/data_batch_{i}" for i in range(1, 6)] + [
        f"{path}/test_batch"
    ]
    cifar_data = None
    cifar_labels = None
    for filename in batch_filenames:
        batch_data, batch_labels = read_cifar_batch(filename)
        if cifar_data is None:
            cifar_data = batch_data
        else:
            cifar_data = np.concatenate((cifar_data, batch_data))
        if cifar_labels is None:
            cifar_labels = batch_labels
        else:
            cifar_labels = np.concatenate((cifar_labels, batch_labels))

    return cifar_data, cifar_labels


def split_dataset(
    data: np.ndarray,
    labels: np.ndarray,
    split: float,
) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    """Splits the dataset into a training and test set.

    Args:
        data (np.ndarray): The data to split.
        labels (np.ndarray): The labels to split.
        split (float): The fraction of the data to use for training.

    Returns:
        (np.ndarray, np.ndarray, np.ndarray, np.ndarray): A tuple containing the training data, training labels, test data, and test labels.
    """
    # get array of indices from the length of the data
    indices = np.arange(data.shape[0])
    # shuffle the indices around
    np.random.shuffle(indices)
    # match entries to the shuffled indices
    shuffled_data = data[indices]
    suffled_labels = labels[indices]

    split_index = round(split * shuffled_data.shape[0])

    data_train = shuffled_data[:split_index]
    labels_train = suffled_labels[:split_index]
    data_test = shuffled_data[split_index:]
    labels_test = suffled_labels[split_index:]

    return data_train, labels_train, data_test, labels_test
